# -*- coding: utf-8 -*-

from libs.auth_code import AuthCode
from libs.password import Password
from libs.access_token import AccessToken
from libs.request import Request
from libs.response import Response
from libs.connection import Connection

class Client(object):

    def __init__(self, client_id, client_secret, **opts):
        self.id = client_id
        self.secret = client_secret
        self.site = opts.pop('site', '')
        self.opts = { 'authorize_url': '/oauth/authorize',
                      'token_url': '/oauth/token',
                      'token_method': 'POST',
                      'connection_opts': {},
                      'raise_errors': True,
                    }
        self.opts.update(opts)

    def __repr__(self):
        return '<OAuth2 Client>'

    def authorize_url(self, params={}):
        return Connection.build_url(self.site, path=self.opts['authorize_url'], params=params)

    def token_url(self, params={}):
        return Connection.build_url(self.site, path=self.opts['token_url'], params=params)

    def request(self, method, url, **opts):
        url = Connection.build_url(self.site, path=url)
        response = Request(method, url, **opts).request()
        return response

    def get_token(self, **opts):
        response = self.request(self.opts['token_method'], self.token_url(), **opts)
        opts.update(response.parsed)
        return AccessToken.from_hash(self, **opts)

    @property
    def password(self):
        return Password(self)

    @property
    def auth_code(self):
        return AuthCode(self)



# -*- coding: utf-8 -*-
from .base import Base

class AuthCode(Base):

    def __repr__(self):
        return '<OAuth2 AuthCode %s>' % self.client.id

    def authorize_params(self, **params):
        params.update({'response_type': 'code', 'client_id': self.client.id})
        return params
    
    def authorize_url(self, **params):
        params = self.authorize_params(**params)
        return self.client.authorize_url(params)

    def get_token(self, code, **opts):
        params = {'grant_type': 'authorization_code', 'code': code}
        params.update(self.client_params)
        opts.update(params)
        return self.client.get_token(**opts)



# -*- coding: utf-8 -*-
import urlparse
import json

def to_json(txt):
    try:
        return json.loads(txt)
    except ValueError:
        return txt

def to_query(txt):
    qs = urlparse.parse_qsl(txt)
    ret = dict(qs)
    return _check_expires_in(ret)

def to_text(txt):
    return txt

def _check_expires_in(ret):
    expires_in = ret.get('expires_in')
    if expires_in and expires_in.isdigit():
        ret['expires_in'] = int(expires_in)
    return ret

PARSERS = {
        'text': to_text,
        'json': to_json,
        'query': to_query,
        }


CONTENT_TYPES = {
            'text/plain': 'text',
            'text/javascript': 'json',
            'application/json': 'json',
            'application/x-www-form-urlencoded': 'query',
            }

class Response(object):

    def __init__(self, response, **opts):
        response, body = response
        self.body = body
        self.response = response
        self.reason = response.reason
        self.status = response.status
        self.content_type = response.get('content-type')

        options = { 'parse': 'text' }
        options.update(opts)

        self.options = options

    def __repr__(self):
        return '<OAuth2 Response>'

    @property
    def parsed(self):
        format = self.options['parse']
        func = PARSERS.get(format, to_text)
        return func(self.body)



# -*- coding: utf-8 -*-

from .base import Base

class Password(Base):
    
    def authorize_url(self):
        return NotImplementedError('The authorization endpoint is not used in this strategy')

    def get_token(self, username, password, **opts):
        params = {'grant_type': 'password',
                  'username': username,
                  'password': password,
                 } 
        params.update(self.client_params)
        opts.update(params)
        return self.client.get_token(**opts)




# -*- coding: utf-8 -*-
import urllib

class Connection(object):

    def __repr__(self):
        return '<OAuth2 Connection>'

    @classmethod
    def build_url(cls, url, path='', params={}):
        params = urllib.urlencode(params)
        params = '?%s'%params if params else ''
        url = path if path.startswith('http') else '%s%s'%(url, path)
        return '%s%s'%(url, params)



# -*- coding: utf-8 -*-
import urllib
from io import BytesIO
from StringIO import StringIO
from httplib2 import Http

from .response import Response
from .multipart import build_multipart

class Request(object):

    def __init__(self, method, url, **opts):
        self.method = method
        self.url = url
        self.headers = opts.pop('headers', {})
        self.body = opts.pop('body', None)

        self.disable_ssl = opts.pop('disable_ssl', True)
        self.http = Http(disable_ssl_certificate_validation=self.disable_ssl)
        self.content_type = 'application/x-www-form-urlencoded'
        self.opts = opts

    def __repr__(self):
        return '<OAuth2 Request>'

    def request(self):
        parse = self.opts.pop('parse', 'json')
        files = self.opts.pop('files', {})
        params = urllib.urlencode(self.opts)

        if self.method in ('POST', 'PUT'):
            (body, content_type) = self.__encode_files(files, self.opts) if files else (params, self.content_type)
            self.headers.update({'Content-Type': content_type})
            self.body = body

        elif self.opts:
            self.url += '&%s'%params if '?' in self.url else '?%s'%params

        response = self.send()
        response = Response(response, parse=parse)

        status = response.status
        #TODO raise error
        if status in (301, 302, 303, 307):
            return response
        elif 200 <= status < 400:
            return response
        elif 400 <= status < 500:
            return response
        elif 500 <= status < 600:
            return response
        return response

    def send(self):
        return self.http.request(self.url, self.method, body=self.body, headers=self.headers)

    def __encode_files(self, files, params):
        if not files or isinstance(files, str):
            return None

        fields = []
        for k, v in tuples(files):
            if isinstance(v, (tuple, list)):
                fn, fp = v
            else:
                fn = guess_filename(v) or k
                fp = v
            if isinstance(fp, str):
                fp = StringIO(fp)
            if isinstance(fp, bytes):
                fp = BytesIO(fp)

            fields.append((k, (fn, fp.read())))

        for k, vs in tuples(params):
            if isinstance(vs, list):
                for v in vs:
                    fields.append((k, str(v)))
            else:
                fields.append((k, str(vs)))

        body, content_type = build_multipart(fields)

        return body, content_type


def guess_filename(obj):
    name = getattr(obj, 'name', None)
    if all([name, not name.startswith('<'), not name.endswith('>')]):
        return name

def tuples(obj):
    if isinstance(obj, dict):
        return list(obj.items())
    elif hasattr(obj, '__iter__'):
        try:
            dict(obj)
        except ValueError:
            pass
        else:
            return obj
    raise ValueError('not a dict or a list of 2-tuples required')



import codecs
import mimetypes
from uuid import uuid4
from io import BytesIO

CRLF = '\r\n'
writer = codecs.lookup('utf-8')[3]

def guess_type(filename):
    return mimetypes.guess_type(filename)[0] or 'application/sctet-stream'

def iter_fields(fields):
    if isinstance(fields, dict):
        fields = fields.iteritems()
    return ((k, v) for k, v in fields)

def build_multipart(fields, boundary=None):
    body = BytesIO()
    boundary = boundary or uuid4().hex

    for field_name, value in iter_fields(fields):
        body.write('--%s%s'%(boundary, CRLF))

        if isinstance(value, tuple):
            file_name, data = value
            writer(body).write('Content-Disposition: form-data; name="%s"; '
                              'filename="%s"%s'%(field_name, file_name, CRLF))
            body.write('Content-Type: %s%s'%(guess_type(file_name), CRLF*2))
        else:
            data = value
            writer(body).write('Content-Disposition: form-data; name="%s"%s'
                               %(field_name, CRLF))
            body.write('Content-Type: text/plain%s'%(CRLF*2))

        if isinstance(data, int):
            data = str(data)

        if isinstance(data, unicode):
            writer(body).write(data)
        else:
            body.write(data)

        body.write(CRLF)

    body.write('--%s--%s'%(boundary, CRLF))
    content_type = 'multipart/form-data; boundary=%s' % boundary
    return body.getvalue(), content_type



# -*- coding: utf-8 -*

class Base(object):

    def __init__(self, client):
        self.client = client

    @property
    def client_params(self):
        return { 'client_id': self.client.id, 'client_secret': self.client.secret }



# -*- coding: utf-8 -*-
import time
import urlparse

class AccessToken(object):

    def __init__(self, client, token, **opts):
        self.client = client
        self.token = token
   
        [setattr(self, attr, opts.pop(attr)) for attr in ['refresh_token', 'expires_in', 'expires_at'] if opts.has_key(attr)]

        if hasattr(self, 'expires_in') and str(self.expires_in).isdigit():
            self.expires_at = int(time.time()) + int(self.expires_in)

        self.opts = { 'mode': opts.pop('mode', 'header'),
                      'header_format': opts.pop('header_format', 'Bearer %s'),
                      'param_name': opts.pop('param_name', 'bearer_token'),
                    }
        self.params = opts

    def __repr__(self):
        return '<OAuth2 AccessToken>'

    @classmethod
    def from_hash(cls, client, **opts):
        return cls(client, opts.pop('access_token', ''), **opts)

    @classmethod
    def from_kvform(cls, client, kvform):
        opts = dict(urlparse.parse_qsl(kvform))
        return cls(client, opts.pop('access_token', ''), **opts)

    def refresh(self, **opts):
        if not getattr(self, 'refresh_token', None):
            raise 'A refresh_token is not available'

        opts = { 'client_id': self.client.id,
                 'client_secret': self.client.secret,
                 'refresh_token': self.refresh_token,
                 'grant_type': 'refresh_token',
               }
        new_token = self.client.get_token(**opts)
        return new_token

    def request(self, method, url, **opts):
        opts = self.__set_token(**opts)
        return self.client.request(method, url, **opts)

    def get(self, url, **opts):
        return self.request('GET', url, **opts)

    def post(self, url, **opts):
        return self.request('POST', url, **opts)

    def put(self, url, **opts):
        return self.request('PUT', url, **opts)

    def patch(self, url, **opts):
        return self.request('PATCH', url, **opts)

    def delete(self, url, **opts):
        return self.request('DELETE', url, **opts)

    @property
    def headers(self):
        return {'Authorization': self.opts['header_format'] % self.token}

    def __set_token(self, **opts):
        mode = self.opts['mode']
        if mode == 'header':
            headers = opts.get('headers', {})
            headers.update(self.headers)
            opts['headers'] = headers
        elif mode == 'query':
            params = opts.get('params', {})
            params[self.opts['param_name']] = self.token
            opts['params'] = params
        elif mode == 'body':
            body = opts.get('body', {})
            if isinstance(body, dict):
                opts['body'][self.opts['param_name']] = token
            else:
                opts['body'] += "&%s=%s"%(self.opts['param_name'], self.token)
        else:
            raise "invalid :mode option of %s"%(self.opts['param_name'])

        return opts



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
    name = 'py-oauth2',
    version = '0.0.4',
    keywords = ('Python', 'OAuth', 'OAuth2', 'Douban', 'GitHub', 'Weibo'),
    description = 'A Python wrapper for the OAuth 2.0 specification.',
    long_description = open('README.rst').read(),
    license = 'MIT License',

    url = 'https://github.com/liluo/py-oauth2',
    author = 'liluo',
    author_email = 'i@liluo.org',

    packages = find_packages(),
    include_package_data = True,
    platforms = 'any',
    install_requires = ['httplib2'],
    classifiers = [
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)
